
		
;Hier kann Debugfunktionalitt eingeschaltet werden 
		.equ	DBG_FAT_INIT = 0
		.equ	DBG_DUMP_CLUSTERS = 0
		.equ	DBG_SDC_READBYTE = 0
		
;Pins der SD/MMC Karte (mssen auf PortB liegen!)
		.equ	sdc_cs	  = 7		;default: pb1
		.equ	sdc_datin = 6		;default: pb2
		.equ	sdc_clk   = 5		;default: pb3
		.equ	sdc_datout= 4		;default: pb4
		
;Durch FAT-Routinen vernderte Register:
		.def	arg0 = r1
		.def	arg1 = r2
		.def	arg2 = r3
		.def	arg3 = r4		;Arguments for SDC-commands
		.def	temp = r16
		.def	cnt0 = r17
		.def	cnt1 = r18
		.def	cnt2 = r19
		.def	cnt3 = r20

;Register die absolut nicht anderweitig benutzt werden drfen:
		.def	aktbyte0 = r5
		.def	aktbyte1 = r6		;Speichert die Byteposision innerhalb eines Sektors
		.def	aktsec = r7		;Aktuelle Sektornummer. Benutzt um das Ende eines Clusters zu berprfen
		
;Register zur freien Verfgung: r8-r15, r21-25 + Pointerregister (X/Y/Z)
		;.def	akku = r21
		
		.def	getnextvalue = r23
		.def	nextvalue    = r24		
		
		
		.cseg


;********************************************************************************
;********************************************************************************
;* SPI functions
;********************************************************************************
;********************************************************************************

;** SPI_GetByte() : Temp
;** Modified: cnt0
SPI_GetByte:	ldi	temp, $ff

;** SPI_SendByte(Temp) : Temp
;** Modified: cnt0
SPI_SendByte:	ldi	cnt0, 8
SPI_SendByte0:	
		cbi	portb, sdc_clk	;CLK=0

		clc
		rol	temp 		; MSB will be sent first

		brcs	SPI_SendByte1
		cbi	portb, sdc_datin; Send zero
		rjmp	SPI_SendByte2
SPI_SendByte1:	sbi	portb, sdc_datin; Send one
SPI_SendByte2:	
		sbi	portb, sdc_clk	;CLK=1

		sbic	pinb, sdc_datout; 0 received?
		ori	temp, 1		; 1 received => Set temp.1
		
		dec	cnt0
		brne	SPI_SendByte0
		sbi	portb, sdc_datin;Raise Data acting like a pull-up
		ret
		

;********************************************************************************
;********************************************************************************
;* SDC functions (low level)
;********************************************************************************
;********************************************************************************

;** SDC_CMD( Temp: Command; ) : Temp
;** Modified: Cnt0

		
SDC_Cmd:	push	temp
		sbi	portb, sdc_cs
		rcall	SPI_GetByte
		cbi	portb, sdc_cs
		pop	temp
		ori	temp, $40
		rcall	SPI_SendByte ;Command
		mov	temp, arg3
		rcall	SPI_SendByte ;arg3
		mov	temp, arg2
		rcall	SPI_SendByte ;arg2
		mov	temp, arg1
		rcall	SPI_SendByte ;arg1
		mov	temp, arg0
		rcall	SPI_SendByte ;arg0
		ldi	temp, $95
		rjmp	SPI_SendByte ;CRC
		
;** SDC_GetResp() : Temp;
;** Modified: Cnt0, Cnt1
SDC_GetResp:	ldi	cnt1, 255
SDC_GetResp0:	rcall	SPI_GetByte
		cpi	Temp, $ff
		brne	SDC_GetRespEnd
		dec	cnt1
		brne	SDC_GetResp0
SDC_GetRespEnd:	ret




;********************************************************************************
;********************************************************************************
;* SDC functions (high level)
;********************************************************************************
;********************************************************************************

;** SDC_Init() : Boolean
;** Sorgt fr eine komplette Initialisierung der SD/MMC-Card
;** Gibt "1" in Temp zurck wenn erfolgreich
SDC_Init:	sbi	portb, sdc_cs		;#cs = 1 (not selected)
		sbi	portb, sdc_datin
		cbi	portb, sdc_clk
		sbi	portb, sdc_datout	;Pullup MISO AN!
		cbi	ddrb, sdc_datout	;DATOUT = Input
		sbi	ddrb, sdc_clk		;CLK = OUTPUT
		sbi	ddrb, sdc_datin		;datin = OUTPUT
		sbi	ddrb, sdc_cs		;#cs = OUTPUT
		
		ldi	cnt2, 10		;Retry up to 10 times
SDC_Init2:	sbi	portb, sdc_cs		;#cs = 1 (not selected)
		ldi	cnt1, 10
SDC_Init0:	rcall	SPI_GetByte
		dec	cnt1
		brne	SDC_Init0
		cbi	portb, sdc_cs
SDC_Init1:	clr	arg0
		clr	arg1
		clr	arg2
		clr	arg3
		ldi	temp, $00 ;CMD0
		rcall	SDC_CMD
		rcall	SDC_GetResp
		cpi	temp, $01
		brne	SDC_Init_Retry
		
		ldi	cnt3, 0			;Retry up to 256 times
SDC_Init3:	ldi	temp, $01 ;CMD1		;bring card out of idle state
		rcall	SDC_CMD
		rcall	SDC_GetResp
		cpi	temp, $00
		breq	SDC_Init4
		dec	cnt3
		brne	SDC_Init3		;Retry...
		rjmp	SDC_Init_Retry		;If we are here, we timed out
SDC_Init4:	sec				;Sucess!
		sbi	portb, sdc_cs
		ret
SDC_Init_Retry:	dec	cnt2
		brne	SDC_Init2
		rjmp	SDC_ERR
		
		
;** SDC_Release
;** Sorgt dafr, dass die Karte entfernt werden kann
		
SDC_Release:	cbi	portb, sdc_datout	;Pullup aus
		cbi	ddrb, sdc_cs
		cbi	ddrb, sdc_datin
		cbi	ddrb, sdc_clk
		cbi	ddrb, sdc_datout	;Alles auf Eingang
		ret

;** SDC_ReadStart
;** Liet den durch arg1:arg2:arg3 spezifizierten sektor
SDC_ReadStart:	lsl	arg1
		rol	arg2
		rol	arg3 ;*
		clr	arg0			;512 alignment sicherstellen
		ldi	temp, 18
		rcall	SDC_CMD
		rcall	SDC_GetResp	
		cpi	temp, $00
		brne	SDC_ERR
		clr	aktbyte0
		clr	aktbyte1
		clr	aktsec
		sec		 		;Signalize Success
		ret

;** SDC_ERR
;** Signalisiert durch entfernen des Carry-Bits einen Fehler
SDC_ERRX:	
SDC_ERR:	clc
		sbi	portb, sdc_cs
		ret

;** SDC_ReadByte
;** Liet das nchste Byte aus dem Sektor (und bercksichtigt Sektorbergnge)
;** Aktuelles Byte steht in Register Temp
SDC_ReadByte:	tst	aktbyte0
		brne	SDC_ReadByte0
		sbrs	aktbyte1, 1
		rjmp	SDC_ReadByte0		;berprfen ob Sektorende erreicht (Byte=512)
		rcall	SPI_GetByte
		rcall	SPI_GetByte		;Sektorende erreicht => CRC auslesen und verwerfen
		clr	aktbyte0
		clr	aktbyte1
		inc	aktsec
SDC_ReadByte0:	
		tst	aktbyte0
		brne	SDC_ReadByte1
		tst	aktbyte1
		brne	SDC_ReadByte1		;berprfen ob 1. Byte eines Sektors
		
		rcall	SDC_GetResp		;StartToken abwarten
		cpi	temp, $fe		;Token korrekt?
		brne	SDC_ERRX		;Nein -> Fehler signalisieren
SDC_ReadByte1:	
		ldi	temp, 1
		add	aktbyte0, temp
		ldi	temp, 0
		adc	aktbyte1, temp
		rcall	SPI_GetByte
		sec
		ret
		
SDC_SkipBytes:	push	temp
		rcall	SDC_ReadByte
		brcc	SDC_ERR
		pop	temp
		dec	temp
		brne	SDC_SkipBytes
		sec
		ret

;** SDC_ReadStop
;** Stopt das Auslesen der Sektoren
SDC_ReadStop:	ldi	temp, 12
		rcall	SDC_CMD
		rcall	SPI_GetByte
		rcall	SDC_GetResp
		cpi	temp, $00
		brne	SDC_ERR
		rcall	SPI_GetByte
		rcall	SPI_GetByte
		sbi	portb, sdc_cs
		sec
		ret
		

;********************************************************************************
;********************************************************************************
;* FAT16 functions (high level)
;********************************************************************************
;********************************************************************************

;** FAT_Err
;** Signalisiert einen Fehler durch Lschen des Carry flags

FAT_ERR:	clc
		ret
		
;** FAT_Init
;** Initialisiert eine SD-Karte und das FAT-System
;** Sucht automatisch nach der 1. Partition, die eine Fat16-Partition sein mu
		
FAT_Init:	
		rcall	SDC_Init		;SD-Karte initialisieren
			brcc	FAT_ERR		;Fehler beim Initialisieren gefunden
				
		clr	arg0
		clr	arg1
		clr	arg2
		clr	arg3
		rcall	SDC_ReadStart
			brcc	FAT_ERR		;Sektor 0 auslesen
		ldi	temp, $F0
		rcall	SDC_SkipBytes
			brcc	FAT_ERR
		ldi	temp, $CE+4
		rcall	SDC_SkipBytes
			brcc	FAT_ERR		;Jump to Position $1BE+4 (Pointer to Partition table
			
		
		rcall	SDC_ReadByte		;Partitionstyp auslesen
			brcc	FAT_ERR

		cpi	temp, $04
		breq	FAT_is16
		cpi	temp, $06
		breq	FAT_is16
		rjmp	FAT_ERR			;Nur Partitionstypen 4 und 6 erlauben (FAT16!)
		
FAT_is16:
		ldi	temp, 3
		rcall	SDC_SkipBytes		;Offset 5-7 verwerfen
			brcc	FAT_ERR
			
		rcall	SDC_ReadByte		;Read 
		mov	arg1, temp
		rcall	SDC_ReadByte		;Read 
		mov	arg2, temp
		rcall	SDC_ReadByte		;Read 
		mov	arg3, temp
		
		rcall	SDC_ReadStop
			brcc	FAT_ERR		;Das prfen des "Magic 55 aa" spar ich mir mal...
		;HIER: arg1:arg2:arg3 zeigt auf 1. Partitionsektor (Bootsektor)
			
		rcall	SDC_ReadStart		;Read Bootsektor
			brcc	FAT_ERR
			
		ldi	temp, $0D
		rcall	SDC_SkipBytes		
		rcall	SDC_ReadByte
		sts	fatsecpercluster, temp	;Get Sectorcount per Cluster
		
		rcall	SDC_ReadByte
		mov	arg0, temp		;Get Low byte of Reserved Sector Count
		rcall	SDC_ReadByte		;Get Hi byte of Reserved Sector Count
		lsr	arg3
		ror	arg2
		ror	arg1
		add	arg1, arg0
		adc	arg2, temp		;Add ResSecCnt To Partition start sector
		
		sts	fatstartsector, arg1
		sts	fatstartsector+1, arg2
		sts	fatstartsector+2, arg3	;Speichern der Fat-Startadresse
		
		rcall	SDC_ReadByte
		mov	cnt3, temp		;Anzahl der Fats holen
		rcall	SDC_ReadByte
		push	temp
		rcall	SDC_ReadByte
		push	temp			;Anzahl der Eintrge im RootDir holen
		
		ldi	temp, $03
		rcall	SDC_SkipBytes
		rcall	SDC_ReadByte		;Low-Byte von SecPerFat
		mov	cnt2, temp
		rcall	SDC_ReadByte		;Hi-Byte von SecPerFat
FAT_INIT0:	add	arg1, cnt2
		adc	arg2, temp
		dec	cnt3
		brne	FAT_INIT0		;NumFats*SecPerFat hinzuaddieren
		sts	fatrootdirsector, arg1
		sts	fatrootdirsector+1, arg2
		sts	fatrootdirsector+2, arg3
		pop	cnt1
		pop	cnt0
			lsr	cnt1
			ror	cnt0
			lsr	cnt1
			ror	cnt0
			lsr	cnt1
			ror	cnt0
			lsr	cnt1
			ror	cnt0		;DirCount /= 16 (*32/512);		
		add	arg1, cnt0
		adc	arg2, cnt1
		ldi	cnt0, 0
		adc	arg3, cnt0
		sts	fatvolumesector, arg1
		sts	fatvolumesector+1, arg2
		sts	fatvolumesector+2, arg3
		rcall	SDC_ReadStop
		sec
		ret
		
;** FAT_ReadStart
;** Initialisiert das Lesen von dem Cluster der durch
;** arg1:arg2 spezifiziert wird
FAT_ReadStart:	
		push	arg1
		push	arg2
		ldi	temp, 2
		sub	arg1, temp
		ldi	temp, 0
		sbc	arg2, temp		;Clusternummer muss um 2 dekr. werden!!
		clr	arg3
		lds	cnt3, fatsecpercluster
FAT_ReadStart1:	cpi	cnt3, 1
		breq	FAT_ReadStart0		;Nur 1 Sektor pro Cluster
		lsl	arg1
		rol	arg2
		rol	arg3
		lsr	cnt3
		rjmp	FAT_ReadStart1
FAT_ReadStart0:	lds	temp, fatvolumesector
		add	arg1, temp
		lds	temp, fatvolumesector+1
		adc	arg2, temp
		lds	temp, fatvolumesector+2
		adc	arg3, temp		;Cluster in Sektornummer umrechnen
		rcall	SDC_ReadStart
		pop	arg2
		pop	arg1
		ret
		
;** FAT_ReadByte
;** Liet ein Byte und springt automatisch zum nchsten Cluster wenn ntig
;** Aktueller Cluster *MUSS* in arg1:arg2 stehen!!
;** Wenn der letzte Cluster gelesen wird, ist das Carry-Flag gelscht, sonst ist es gesetzt
FAT_ReadByte:	lds	temp, fatsecpercluster
		cp	aktsec, temp		;befinden wir uns im letzten Sektor?
		brne	FAT_ReadByte0
			rcall	SDC_ReadStop
			mov	cnt3, arg1	;arg1 sichern
			mov	cnt1, arg2	;arg2 sichern
			lds	arg1, fatstartsector
			add	arg1, cnt1
			lds	arg2, fatstartsector+1
			ldi	cnt1, 0
			adc	arg2,cnt1
			lds	arg3, fatstartsector+2
			adc	arg3,cnt1	;Sektornummer des FAT-Cluster berechnet

			rcall	SDC_ReadStart
			
FAT_ReadByte1:		cpi	cnt3, 0
			breq	FAT_ReadByte2
			rcall	SDC_ReadByte			
			rcall	SDC_ReadByte
			dec	cnt3		;Sektor-Offset dekrementieren
			rjmp	FAT_ReadByte1
			
			
FAT_ReadByte2:	
			rcall	SDC_ReadByte
			mov	arg1, temp
			rcall	SDC_ReadByte
			mov	arg2, temp	;FAT-Eintrag lesen				
			rcall	SDC_ReadStop
			

			ldi	temp, $FF
			cp	arg2, temp
			brne	FAT_ReadByte3
			
			mov	temp, arg1
			andi	temp, $F8
			cpi	temp, $F8
			brne	FAT_ReadByte3
			clc			;Hier: ENDE der Datei erreicht
			ret

FAT_ReadByte3:		rcall	FAT_ReadStart
rcall SDC_ReadByte
			
FAT_ReadByte0:	rjmp	SDC_ReadByte

		
;** FAT_ReadStop
;** Beendet Lesevorgang der Datei/Directory
FAT_ReadStop:	rjmp	SDC_ReadStop


;** FAT_FindFirst
;** Initializes reading directory entries

FAT_FindFirst:	lds	arg1, fatrootdirsector
		lds	arg2, fatrootdirsector+1
		lds	arg3, fatrootdirsector+2
		rjmp	SDC_ReadStart
		
;** FAT_FindNext
;** Reads a Directory Entry
;** if byte at [fatdirentry] equals 0 then the last entry has been read
;** Der Cluster des aktuellen Eintrages wird nach arg1:arg2 bertragen
FAT_FindNext:	push	xl
		push	xh
FAT_FindNext2:	ldi	xl, low(fatdirentry)
		ldi	xh, high(fatdirentry)
		ldi	cnt3, 11
FAT_FindNext0:	rcall	SDC_ReadByte
		st	x+, temp
		dec	cnt3
		brne	FAT_FindNext0		;Verzeichniseintrag ist kopier
		rcall	SDC_ReadByte		;Attribute lesen
		mov	cnt3, temp		;Attribut nach cnt3 bertragen
		sts	fatdirentrytype, temp	;Attribut im Ram speichern
			ldi	temp, 14
			rcall	SDC_SkipBytes	;Zur Clusternummer fortschreiten
		rcall	SDC_ReadByte
		mov	arg1, temp
		rcall	SDC_ReadByte
		mov	arg2, temp
			ldi	temp, 4
			rcall	SDC_SkipBytes
		lds	temp, fatdirentry
		cpi	temp, $00
		breq	FAT_FindNext1		;Letzter Eintrag
		cpi	temp, $E5
		breq	FAT_FindNext2		;Gelschte Datei -> Nchster Eintrag
		andi	cnt3, $08		;Verzeichnisse und Volumenamen berspringen
		cpi	cnt3, 0
		brne	FAT_FindNext2
FAT_FindNext1:	pop	xh
		pop	xl
		ret
		
		
;** FAT_FindClose
;** Will stop the reading of a Directory

FAT_FindClose:	push	arg1
		push	arg2	
		rcall	SDC_ReadStop
		pop	arg2
		pop	arg1
		ret		
		
		









;****************************************************************************
;** DATENSEGMENT - Speichert unter anderem Informationen ber das Dateisystem
;****************************************************************************

			.dseg
fatsecpercluster:	.BYTE	1
fatstartsector:		.BYTE   3	;Startsektor der File Allocation Table
fatrootdirsector:	.BYTE	3	;Startsektor der Root-Directory
fatvolumesector:	.BYTE	3	;Erster Sektor der volume (Datenbereich nach dem Hauptverzeichnis)
fatdirentry:		.BYTE	11	;Ein Dateiname (ohne Punkt - Roh - Leerstellen mit Leerzeichen aufgefllt)
fatdirentrytype:	.BYTE	1	;enthlt Dateiattribut



;DBG_FAT_INIT:
;    0.Phase
;00: Dump Partition Type
;01: Bootsec 0. Byte of Sector
;02:   Bootsec 1. Byte of Sector
;03:   Bootsec 2. Byte of Sector
;04: SecPerCluster
;05: ResSecCnt 0. Byte
;06:   ResSecCnt 0. Byte
;07: FatStartSector 0. Byte
;08:   FatStartSector 1. Byte
;09:   FatStartSector 2. Byte
;10: FatCnt
;11: RootDirCnt 0.Byte
;12:   RootDirCnt 1.Byte
;13: SecPerFat 0. Byte
;14:   SecPerFat 1. Byte
;15: Startsector of MainDirArea 0. Byte
;16:   Startsector of MainDirArea 1. Byte
;17:   Startsector of MainDirArea 2. Byte
;18: Startsector of data area 0. Byte
;19:   Startsector of data area 1. Byte
;20:   Startsector of data area 2. Byte

;DBG_DUMP_CLUSTERS:
;    1.Phase (ReadStart)
;00: Cluster Number 0. Byte
;01:   Cluster Number 1. Byte
;02: Fat Sector Number 0. Byte
;03:   Fat Sector Number 1. Byte
;04:   Fat Sector Number 2. Byte

;DBG_DUMP_CLUSTERS:
;    2.Phase (ReadByte)
;00: Cluster Number 0. Byte
;01:   Cluster Number 1. Byte


;DBG_SDC_READBYTE:
;    3.Phase (ReadByte)
;00: Start-Token-Response (ideally $FE)
